@@ -33,7 +33,7 @@
 #define DEFAULT_MAX_RUNTIME                                                                        \
 	DatumGetIntervalP(DirectFunctionCall3(interval_in, CStringGetDatum("0"), InvalidOid, -1))
 /* Default buckets per batch is 1, which means that the job will refresh 1 bucket at a time */
-#define DEFAULT_BUCKETS_PER_BATCH 1
+#define DEFAULT_BUCKETS_PER_BATCH 10
 /* Default max batches per execution is 0, which means no limit */
 #define DEFAULT_MAX_BATCHES_PER_EXECUTION 0
 /* Default refresh newest first is true, which means from newest data to the oldest */
@@ -71,7 +71,7 @@ get_time_from_interval(const Dimension *dim, Datum interval, Oid type)
 	}
 	else if (type == INTERVALOID)
 	{
-		Datum res = subtract_interval_from_now(DatumGetIntervalP(interval), partitioning_type);
+		Datum res = ts_subtract_interval_from_now(DatumGetIntervalP(interval), partitioning_type);
 		return ts_time_value_to_internal(res, partitioning_type);
 	}
 	else
