@@ -47,6 +47,7 @@
 #include "hypercube.h"
 #include "hypertable.h"
 #include "hypertable_cache.h"
+#include "nodes/columnar_scan/columnar_scan.h"
 #include "recompress.h"
 #include "scan_iterator.h"
 #include "scanner.h"
@@ -664,7 +665,6 @@ static bool
 recompress_chunk_impl(Chunk *chunk, Oid *uncompressed_chunk_id, bool recompress)
 {
 	CompressionSettings *chunk_settings = ts_compression_settings_get(chunk->table_id);
-	CompressionSettings *ht_settings = ts_compression_settings_get(chunk->hypertable_relid);
 	bool recompressed = false;
 
 	if (!chunk_settings || !chunk_settings->fd.orderby)
@@ -689,7 +689,7 @@ recompress_chunk_impl(Chunk *chunk, Oid *uncompressed_chunk_id, bool recompress)
 	/*
 	 * Choose recompression strategy based on chunk state and settings:
 	 * 1. Segmentwise recompression: For partial chunks
-	 * 2. In-memory recompression: For fully compressed chunks with matching settings
+	 * 2. In-memory recompression: For other compressed chunks with matching settings
 	 * 3. Fallback to decompress/compress: When neither strategy is applicable
 	 */
 
@@ -707,8 +707,7 @@ recompress_chunk_impl(Chunk *chunk, Oid *uncompressed_chunk_id, bool recompress)
 		*uncompressed_chunk_id = recompress_chunk_segmentwise_impl(chunk);
 		recompressed = true;
 	}
-	/* TODO: optimize cases where settings differ but recompression is still possible */
-	else if (recompress && ts_compression_settings_equal_with_defaults(ht_settings, chunk_settings))
+	else
 	{
 		if (!ts_guc_enable_in_memory_recompression)
 		{
@@ -900,6 +899,70 @@ tsl_decompress_chunk(PG_FUNCTION_ARGS)
 	PG_RETURN_OID(uncompressed_chunk_id);
 }
 
+static bool
+can_use_in_memory_rebuild(Chunk *chunk)
+{
+	CompressionSettings *chunk_settings = ts_compression_settings_get(chunk->table_id);
+
+	/* check if we can allow in-memory recompression to rebuild columnstore */
+	if (!chunk_settings || !chunk_settings->fd.orderby)
+	{
+		elog(DEBUG1, "in-memory rebuild columnstore is disabled due to no order by");
+		return false;
+	}
+	if (!get_compressed_chunk_index_for_recompression(chunk))
+	{
+		elog(DEBUG1, "in-memory rebuild columnstore is disabled due to no compressed chunk index.");
+		return false;
+	}
+	if (!ts_guc_enable_in_memory_recompression)
+	{
+		elog(DEBUG1, "timescaledb.enable_in_memory_recompression is disabled");
+		return false;
+	}
+
+	return true;
+}
+
+Datum
+tsl_rebuild_columnstore(PG_FUNCTION_ARGS)
+{
+	Oid chunk_relid = PG_ARGISNULL(0) ? InvalidOid : PG_GETARG_OID(0);
+
+	ts_feature_flag_check(FEATURE_HYPERTABLE_COMPRESSION);
+
+	TS_PREVENT_FUNC_IF_READ_ONLY();
+
+	if (!OidIsValid(chunk_relid))
+		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("invalid chunk OID")));
+
+	Chunk *chunk = ts_chunk_get_by_relid(chunk_relid, true);
+
+	if (!ts_chunk_is_compressed(chunk) || ts_chunk_is_frozen(chunk))
+	{
+		ereport(NOTICE,
+				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+				 errmsg("chunk \"%s.%s\" is uncompressed or frozen, skipping",
+						NameStr(chunk->fd.schema_name),
+						NameStr(chunk->fd.table_name))));
+		PG_RETURN_VOID();
+	}
+
+	/* Try rebuild with in-memory recompression, fall back to decompress/compress if needed */
+	if (!can_use_in_memory_rebuild(chunk) || !recompress_chunk_in_memory_impl(chunk))
+	{
+		elog(DEBUG1,
+			 "falling back to decompress/compress, performing full "
+			 "rebuild on chunk \"%s.%s\"",
+			 NameStr(chunk->fd.schema_name),
+			 NameStr(chunk->fd.table_name));
+		decompress_chunk_impl(chunk, false);
+		compress_chunk_impl(chunk->hypertable_relid, chunk->table_id);
+	}
+
+	PG_RETURN_VOID();
+}
+
 /*
  * This is hacky but it doesn't matter. We just want to check for the existence of such an index
  * on the compressed chunk.
@@ -946,3 +1009,15 @@ tsl_compression_chunk_create(Hypertable *compressed_ht, Chunk *src_chunk)
 	/* Create a new compressed chunk */
 	return create_compress_chunk(compressed_ht, src_chunk, InvalidOid);
 }
+
+Datum
+tsl_estimate_compressed_batch_size(PG_FUNCTION_ARGS)
+{
+	Oid relid = PG_GETARG_OID(0);
+
+	ts_feature_flag_check(FEATURE_HYPERTABLE_COMPRESSION);
+
+	float8 approx_batch_size = (float8) ts_columnar_estimate_compressed_batch_size(relid);
+
+	PG_RETURN_FLOAT8(approx_batch_size);
+}
