@@ -20,8 +20,10 @@
 #include <parser/parse_relation.h>
 #include <parser/parsetree.h>
 #include <planner/planner.h>
+#include <storage/lockdefs.h>
 #include <utils/builtins.h>
 #include <utils/lsyscache.h>
+#include <utils/syscache.h>
 #include <utils/typcache.h>
 
 #include <planner.h>
@@ -34,6 +36,7 @@
 #include "debug_assert.h"
 #include "import/allpaths.h"
 #include "import/planner.h"
+#include "nodes/columnar_index_scan/columnar_index_scan.h"
 #include "nodes/columnar_scan/columnar_scan.h"
 #include "nodes/columnar_scan/planner.h"
 #include "nodes/columnar_scan/qual_pushdown.h"
@@ -626,6 +629,8 @@ build_compressioninfo(PlannerInfo *root, const Hypertable *ht, const Chunk *chun
 		info->parent_relids = find_childrel_parents(root, chunk_rel);
 	}
 
+	info->chunk_status = chunk->fd.status;
+
 	return info;
 }
 
@@ -634,22 +639,20 @@ build_compressioninfo(PlannerInfo *root, const Hypertable *ht, const Chunk *chun
  * Postgres statistics for _ts_meta_count column.
  * Returns TARGET_COMPRESSED_BATCH_SIZE when no pg_statistic entry exists.
  */
-static double
-estimate_compressed_batch_size(PlannerInfo *root, const CompressionInfo *compression_info)
+double
+ts_columnar_estimate_compressed_batch_size(const Oid relid)
 {
-	AttrNumber attnum = get_attnum(compression_info->compressed_rte->relid, "_ts_meta_count");
+	AttrNumber attnum = get_attnum(relid, "_ts_meta_count");
 	if (attnum == InvalidAttrNumber)
 		return TARGET_COMPRESSED_BATCH_SIZE;
 
-	Var *var = makeVar(compression_info->compressed_rel->relid, attnum, INT4OID, -1, InvalidOid, 0);
-
 	/* fetch statistics */
-	VariableStatData vardata;
-	examine_variable(root, (Node *) var, 0, &vardata);
-
-	if (!HeapTupleIsValid(vardata.statsTuple))
+	HeapTuple statsTuple = SearchSysCache3(STATRELATTINH,
+										   ObjectIdGetDatum(relid),
+										   Int16GetDatum(attnum),
+										   BoolGetDatum(false));
+	if (!HeapTupleIsValid(statsTuple))
 	{
-		ReleaseVariableStats(vardata);
 		return TARGET_COMPRESSED_BATCH_SIZE;
 	}
 
@@ -659,7 +662,7 @@ estimate_compressed_batch_size(PlannerInfo *root, const CompressionInfo *compres
 	/* exact MCV contribution */
 	AttStatsSlot mcvslot;
 	if (get_attstatsslot(&mcvslot,
-						 vardata.statsTuple,
+						 statsTuple,
 						 STATISTIC_KIND_MCV,
 						 InvalidOid,
 						 ATTSTATSSLOT_VALUES | ATTSTATSSLOT_NUMBERS))
@@ -679,7 +682,7 @@ estimate_compressed_batch_size(PlannerInfo *root, const CompressionInfo *compres
 	/* histogram contribution */
 	AttStatsSlot histslot;
 	if (get_attstatsslot(&histslot,
-						 vardata.statsTuple,
+						 statsTuple,
 						 STATISTIC_KIND_HISTOGRAM,
 						 InvalidOid,
 						 ATTSTATSSLOT_VALUES))
@@ -699,7 +702,7 @@ estimate_compressed_batch_size(PlannerInfo *root, const CompressionInfo *compres
 		free_attstatsslot(&histslot);
 	}
 
-	ReleaseVariableStats(vardata);
+	ReleaseSysCache(statsTuple);
 
 	const double final_result = mcv_sum + hist_sum;
 	if (final_result == 0)
@@ -1138,7 +1141,7 @@ ts_columnar_scan_generate_paths(PlannerInfo *root, RelOptInfo *chunk_rel, const
 	 * statistics.
 	 */
 	compression_info->compressed_batch_size =
-		estimate_compressed_batch_size(root, compression_info);
+		ts_columnar_estimate_compressed_batch_size(compression_info->compressed_rte->relid);
 
 	/*
 	 * Estimate the size of decompressed chunk based on the compressed chunk.
@@ -1361,6 +1364,22 @@ build_on_single_compressed_path(PlannerInfo *root, const Chunk *chunk, RelOptInf
 						"time")));
 	}
 
+	/*
+	 * Try to create a ColumnarIndexScan path for the metadata-only optimization.
+	 */
+	{
+		ColumnarIndexScanPath *index_scan_path =
+			ts_columnar_index_scan_path_create(root,
+											   chunk,
+											   chunk_rel,
+											   compression_info,
+											   compressed_path);
+		if (index_scan_path)
+		{
+			decompressed_paths = lappend(decompressed_paths, index_scan_path);
+		}
+	}
+
 	/*
 	 * If we can push down the sort below the ColumnarScan node, we set the
 	 * pathkeys of the decompress node to the decompressed_sort_pathkeys. We
@@ -2048,6 +2067,7 @@ add_segmentby_to_equivalence_class(PlannerInfo *root, EquivalenceClass *cur_ec,
 	{
 		cur_em = (EquivalenceMember *) lfirst(lc);
 #endif
+		Node *node;
 		Expr *child_expr;
 		Relids new_relids;
 		Var *var;
@@ -2055,11 +2075,10 @@ add_segmentby_to_equivalence_class(PlannerInfo *root, EquivalenceClass *cur_ec,
 
 		/* only consider EquivalenceMembers that are Vars, possibly with RelabelType, of the
 		 * uncompressed chunk */
-		var = (Var *) cur_em->em_expr;
-		while (var && IsA(var, RelabelType))
-			var = (Var *) ((RelabelType *) var)->arg;
-		if (!(var && IsA(var, Var)))
+		node = strip_implicit_coercions((Node *) cur_em->em_expr);
+		if (!(node && IsA(node, Var)))
 			continue;
+		var = castNode(Var, node);
 
 		/*
 		 * We want to base our equivalence member on the hypertable equivalence
@@ -2233,8 +2252,14 @@ columnar_scan_add_plannerinfo(PlannerInfo *root, CompressionInfo *info, const Ch
 	/*
 	 * Add the compressed chunk to the baserel cache. Note that it belongs to
 	 * a different hypertable, the internal compression table.
+	 *
+	 * Ensure we do not grab a slice lock because that will assign a transaction ID that could
+	 * unnecessarily block other operations.
 	 */
-	const Chunk *compressed_chunk = ts_chunk_get_by_relid(info->settings->fd.compress_relid, true);
+	const Chunk *compressed_chunk = ts_chunk_get_by_relid_locked(info->settings->fd.compress_relid,
+																 AccessShareLock,
+																 NULL,
+																 true);
 	ts_add_baserel_cache_entry_for_chunk(info->settings->fd.compress_relid,
 										 ts_planner_get_hypertable(compressed_chunk
 																	   ->hypertable_relid,
@@ -2370,6 +2395,7 @@ columnar_scan_path_create(PlannerInfo *root, const CompressionInfo *compression_
 
 	path->custom_path.custom_paths = list_make1(compressed_path);
 	path->reverse = false;
+	path->chunk_status = compression_info->chunk_status;
 	path->required_compressed_pathkeys = NIL;
 	cost_columnar_scan(root, compression_info, &path->custom_path.path, compressed_path);
 
@@ -2552,29 +2578,25 @@ find_const_segmentby(RelOptInfo *chunk_rel, const CompressionInfo *info)
 			if (IsA(ri->clause, OpExpr) && list_length(castNode(OpExpr, ri->clause)->args) == 2)
 			{
 				OpExpr *op = castNode(OpExpr, ri->clause);
-				Var *lvar, *rvar, *var;
+				Node *lnode, *rnode;
+				Var *var;
 				Expr *other;
 
 				if (op->opretset)
 					continue;
 
-				lvar = linitial(op->args);
-				while (lvar && IsA(lvar, RelabelType))
-					lvar = (Var *) ((RelabelType *) lvar)->arg;
+				lnode = strip_implicit_coercions(linitial(op->args));
+				rnode = strip_implicit_coercions(lsecond(op->args));
 
-				rvar = lsecond(op->args);
-				while (rvar && IsA(rvar, RelabelType))
-					rvar = (Var *) ((RelabelType *) rvar)->arg;
-
-				Assert(lvar && rvar);
-				if (IsA(lvar, Var))
+				Assert(lnode && rnode);
+				if (IsA(lnode, Var))
 				{
-					var = castNode(Var, lvar);
+					var = castNode(Var, lnode);
 					other = lsecond(op->args);
 				}
-				else if (IsA(rvar, Var))
+				else if (IsA(rnode, Var))
 				{
-					var = castNode(Var, rvar);
+					var = castNode(Var, rnode);
 					other = linitial(op->args);
 				}
 				else
@@ -2588,7 +2610,41 @@ find_const_segmentby(RelOptInfo *chunk_rel, const CompressionInfo *info)
 					TypeCacheEntry *tce = lookup_type_cache(var->vartype, TYPECACHE_EQ_OPR);
 
 					if (op->opno != tce->eq_opr)
-						continue;
+					{
+						/* Issue #9066: check if our OpExpr is still an equality (Var = Const)
+						 * for Var and Const of different types
+						 */
+#if PG18_GE
+						List *opinfos = get_op_index_interpretation(op->opno);
+#else
+						List *opinfos = get_op_btree_interpretation(op->opno);
+#endif
+						ListCell *lc;
+						bool equality = false;
+						foreach (lc, opinfos)
+						{
+#if PG18_GE
+							OpIndexInterpretation *opinfo = (OpIndexInterpretation *) lfirst(lc);
+							if (opinfo->cmptype == COMPARE_EQ)
+#else
+							OpBtreeInterpretation *opinfo = (OpBtreeInterpretation *) lfirst(lc);
+							if (opinfo->strategy == BTEqualStrategyNumber)
+#endif
+							{
+								Oid mixed_type_eqop = get_opfamily_member(opinfo->opfamily_id,
+																		  var->vartype,
+																		  exprType((Node *) other),
+																		  BTEqualStrategyNumber);
+								if (op->opno == mixed_type_eqop)
+								{
+									equality = true;
+									break;
+								}
+							}
+						}
+						if (!equality)
+							continue;
+					}
 
 					if (bms_is_member(var->varattno, info->chunk_segmentby_attnos))
 						segmentby_columns = bms_add_member(segmentby_columns, var->varattno);
@@ -2614,16 +2670,14 @@ match_pathkeys_to_compression_orderby(List *pathkeys, List *chunk_em_exprs,
 	{
 		compressed_pk_index++;
 		PathKey *pk = list_nth_node(PathKey, pathkeys, i);
-		Expr *expr = (Expr *) list_nth(chunk_em_exprs, i);
-		while (expr && IsA(expr, RelabelType))
-			expr = ((RelabelType *) expr)->arg;
+		Node *node = strip_implicit_coercions((Node *) list_nth(chunk_em_exprs, i));
 
-		if (expr == NULL || !IsA(expr, Var))
+		if (node == NULL || !IsA(node, Var))
 		{
 			return false;
 		}
 
-		Var *var = castNode(Var, expr);
+		Var *var = castNode(Var, node);
 
 		if (var->varattno <= 0)
 		{
@@ -2698,7 +2752,7 @@ match_pathkeys_to_compression_orderby(List *pathkeys, List *chunk_em_exprs,
 }
 
 /*
- * Check if we can push down the sort below the ColumnarSacn node and fill
+ * Check if we can push down the sort below the ColumnarScan node and fill
  * SortInfo accordingly
  *
  * The following conditions need to be true for pushdown:
@@ -2765,12 +2819,13 @@ build_sortinfo(PlannerInfo *root, const Chunk *chunk, RelOptInfo *chunk_rel,
 	cost_qual_eval(&sort_info.decompressed_sort_pathkeys_cost, sort_pathkey_exprs, root);
 
 	/*
-	 * Next, check if we can push the sort down to the uncompressed part.
+	 * Next, check if we can push the sort down to the compressed part.
 	 *
-	 * Not possible if the chunk is unordered.
+	 * Batch sorted merge optimization is enabled for unordered chunks
+	 * because we do merge the batches at execution time which
+	 * only relies that the batches themselves are sorted which is
+	 * always the case.
 	 */
-	if (ts_chunk_is_unordered(chunk))
-		return sort_info;
 
 	/* all segmentby columns need to be prefix of pathkeys */
 	int i = 0;
@@ -2793,13 +2848,11 @@ build_sortinfo(PlannerInfo *root, const Chunk *chunk, RelOptInfo *chunk_rel,
 		{
 			Assert(bms_num_members(segmentby_columns) <= compression_info->num_segmentby_columns);
 
-			Expr *expr = (Expr *) list_nth(chunk_em_exprs, i);
-			while (expr && IsA(expr, RelabelType))
-				expr = ((RelabelType *) expr)->arg;
+			Node *node = strip_implicit_coercions((Node *) list_nth(chunk_em_exprs, i));
 
-			if (expr == NULL || !IsA(expr, Var))
+			if (node == NULL || !IsA(node, Var))
 				break;
-			var = castNode(Var, expr);
+			var = castNode(Var, node);
 
 			if (var->varattno <= 0)
 				break;
@@ -2811,6 +2864,16 @@ build_sortinfo(PlannerInfo *root, const Chunk *chunk, RelOptInfo *chunk_rel,
 			segmentby_columns = bms_add_member(segmentby_columns, var->varattno);
 		}
 
+		/*
+		 * Pathkeys satisfied by sorting the compressed data on segmentby columns.
+		 * Can use compressed sort on segmentby cols for unordered chunks as well.
+		 */
+		if (i == list_length(pathkeys))
+		{
+			sort_info.use_compressed_sort = true;
+			return sort_info;
+		}
+
 		/*
 		 * If pathkeys still has items, but we didn't find all segmentby columns,
 		 * we cannot satisfy these pathkeys by sorting the compressed chunk table.
@@ -2835,14 +2898,12 @@ build_sortinfo(PlannerInfo *root, const Chunk *chunk, RelOptInfo *chunk_rel,
 		}
 	}
 
-	if (i == list_length(pathkeys))
-	{
-		/*
-		 * Pathkeys satisfied by sorting the compressed data on segmentby columns.
-		 */
-		sort_info.use_compressed_sort = true;
+	/*
+	 * Cannot push down sort on non-segmentby columns
+	 * if the chunk has batches overlapping on orderby columns
+	 */
+	if (ts_chunk_is_unordered(chunk))
 		return sort_info;
-	}
 
 	/*
 	 * Pathkeys includes columns past segmentby columns, so we need sequence_num
